---
title: A Walkthrough of the Interactive Graph Editor (WIP)
---

## Introductions and Imports

Recently I released version `0.1.0.0` of [ige](https://github.com/olynch/ige), and I thought I would try and break down the code in order to show some of the key ideas behind it. If you don't know Haskell, this will probably not be that interesting, but I will be flattered if you persevere anyways.

As a way of figuring out where identifiers in code samples come from, let the following import block serve as a reference. Identifiers without qualifications are either defined elsewhere in my own code or are part of [Protolude](https://github.com/sdiehl/protolude).

```haskell
{-# LANGUAGE NoImplicitPrelude #-}

import Protolude
import qualified Conduit as C
import qualified Data.Graph.Inductive.PatriciaTree as PT
import qualified Graphics.UI.Gtk as Gtk
import qualified Control.Concurrent.STM as STM
import qualified Data.Complex as Complex
import qualified Data.Map.Strict as Map
```
## Ideas

The most interesting and original part of IGE is the way interaction with the user is expressed. In order to support different keybindings at different times (ie., when the user is typing a command, letters should not trigger actions, they should instead be added to the command string), I needed some sort of state. One way would be to have an explicit mode variable, and present a different set of keybinding based on the current value of the mode variable. However, then every time I wanted to add a new mode, I would have to change the type of the state variable, which is not only annoying but also difficult to extend.

In order to create a more flexible solution, I represent the current mode in IGE by which function I am in. Each function typically begins by waiting for a keypress, and then does some action and then either recursively calls itself or calls another function to switch into a different mode. Adding a new mode is a simple as writing a function and then editing one of the existing functions to call that function.

That all sounds dandy, but I'm sure that when you all saw "waiting for a keypress" you thought "sounds imperative and side-effecty but OK". Well, never fear: there's a monad for that! If you are not familiar with the excellent [conduit](https://hackage.haskell.org/package/conduit) library, I suggest you take a look at [the tutorial](https://haskell-lang.org/library/conduit).

## Core Types

The core of IGE is the KeyBinding type. It is defined by

```haskell
type KeyBinding n e = C.ConduitM Gtk.KeyVal RefreshType (IGEM n e)
```

At a high level, what the KeyBinding type encodes is the ability to receive keystrokes, change the editor state, send refresh commands to the code that displays the editor window, and perform arbitrary IO actions. The `ConduitM KeyVal RefreshType` provides the ability to receive keystrokes and send refresh signals. This is stacked on top of IGEM, which is defined by

```haskell
newtype IGEM n e a = IGEM { unIGEM :: ReaderT (STM.TVar (EditorState n e)) IO a }
  deriving (Functor, Applicative, Monad, MonadIO, MonadReader (TVar (EditorState n e)))
```

This follows the [`ReaderT` design pattern](https://www.fpcomplete.com/blog/2017/06/readert-design-pattern), though because all of the fields of EditorState are meant to be mutable, it was easier to put the whole thing in a TVar.
[^1]
At this point, you are probably wondering what the `n` and `e` type variables that keep being shunted along refer to. The answer lies in the definition of EditorState, which is

```haskell
type ℂ = Complex.Complex Double

-- "Rigid Motion"
data RM = RM {
    rmAT :: ℂ -- "Amplitwist"
  -- magnitude is zoom level, argument is rotation

  , rmTrans :: ℂ -- "Translation"
  }

data EditorState n e = EditorState {
    esGraph :: PT.Gr n e

  , esRM :: RM -- rigid motion, 

  , esNum :: Int -- needed to give new nodes unique ids,
  -- incremented every time a new node is added

  , esCommand :: [Char] -- the text shown at the bottom of the screen 
  -- ala vim command line, stored backwards so that appending is natural

  , esLabels :: [([Char], Node)] -- labels to draw on top of nodes
  -- used to select nodes using the keyboard

  , esNodeMap :: Map.Map Node ℂ -- Where each node is placed on the virtual
  -- editing surface. The esRM rigid motion is applied to each value before
  -- the graph is rendered.
  }
```

[^1]: Who knows if STM is overkill for my purposes (I don't have a ton of concurrency), but I wanted to learn the API.
